/* AFC - The  Advanced Foundation Classes
 * Copyright (C) 2000/2004  Fabio Rotondo - Andrea Galimberti
 *  
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mem_tracker.h"

// {{{ docs
/*
@config
	TITLE:     MemTracker
	VERSION:   1.00
	AUTHOR:    Fabio Rotondo - fabio@rotondo.it
@endnode

@node quote
	*Who are you going to believe, me or your own eyes?*

		Groucho Marx
@endnode

@node intro

MemTracker is a class to manage memory allocations and deallocations. It can be used as a resource tracker inside your programs,
allowing you to almost forget about freeing memory. 

If you are writing *AFC compliant* programs, using afc_malloc() and afc_free() each time you need to malloc/free some memory, then
you can implement MemTracker in a breeze. This will assure you that all memory allocations you made in your program will be free()ed
by MemTracker in the right moment your application ends.

MemTracker is a *base class*, that means it does not depends upon other AFC classes.

@endnode
*/
// }}}

static const char class_name[] = "MemTracker";

static int afc_mem_tracker_internal_double_array ( MemTracker * );
static int afc_mem_tracker_internal_sort ( const void * hd1, const void * hd2 );
static void * afc_mem_tracker_internal_find ( MemTracker * mt, void * mem );
static void * afc_mem_tracker_internal_do_sort ( MemTracker * am, int ( *comp ) ( const void *, const void * ) ) ;
static int afc_mem_tracker_internal_del ( MemTracker * am );
static void * afc_mem_tracker_internal_item ( MemTracker * am, unsigned long item );
//static void afc_mem_tracker_internal_debug_dump ( MemTracker * mt );

// {{{ MemTracker * afc_mem_tracker_new ()   
/*
@node afc_mem_tracker_new

	         NAME: afc_mem_tracker_new ()    - Initializes a new afc_mem_tracker instance.

	     SYNOPSIS: MemTracker * afc_mem_tracker_new ( )

	  DESCRIPTION: This function initializes a new afc_mem_tracker instance.

	        INPUT: NONE

	      RESULTS: a valid inizialized afc_mem_tracker structure. NULL in case of errors.

	     SEE ALSO: - afc_mem_tracker_delete()

@endnode
*/
MemTracker * afc_mem_tracker_new ( )
{
TRY ( MemTracker * );

	MemTracker * mem_tracker = malloc ( sizeof ( struct afc_mem_tracker ) );

	if ( mem_tracker == NULL ) RAISE_FAST_RC ( AFC_ERR_NO_MEMORY, "mem_tracker", NULL );
	mem_tracker->magic = AFC_MEM_TRACKER_MAGIC;

	if ( ( mem_tracker->mem = malloc ( ( sizeof ( void * ) ) * AFC_MEM_TRACKER_DEFAULT_ITEMS ) ) == NULL ) RAISE_FAST_RC ( AFC_ERR_NO_MEMORY, "mem", NULL );

	mem_tracker->current_pos = 0;
	mem_tracker->max_items   = AFC_MEM_TRACKER_DEFAULT_ITEMS;

	RETURN ( mem_tracker );

EXCEPT
	afc_mem_tracker_delete ( mem_tracker );

FINALLY

ENDTRY
}
// }}}
// {{{ int afc_mem_tracker_delete ( MemTracker * mem_tracker ) 
/*
@node afc_mem_tracker_delete

	         NAME: afc_mem_tracker_delete ( mem_tracker )  - Disposes a valid mem_tracker instance.

	     SYNOPSIS: int afc_mem_tracker_delete (MemTracker * mem_tracker)

	  DESCRIPTION: This function frees an already alloc'd afc_mem_tracker structure.

	        INPUT: - mem_tracker  - Pointer to a valid afc_mem_tracker class.

	      RESULTS: should be AFC_ERR_NO_ERROR

	        NOTES: - this method calls: afc_mem_tracker_clear()

	     SEE ALSO: - afc_mem_tracker_new()
	               - afc_mem_tracker_clear()
@endnode

*/
int _afc_mem_tracker_delete ( MemTracker * mem_tracker ) 
{
	int afc_res; 

	if ( ( afc_res = afc_mem_tracker_clear ( mem_tracker ) ) != AFC_ERR_NO_ERROR ) return ( afc_res );

	if ( mem_tracker->mem ) free ( mem_tracker->mem );

	/* NOTE: any class contained in afc_mem_tracker should be deleted here */

	free ( mem_tracker );

	return ( AFC_ERR_NO_ERROR );
}
// }}}
// {{{ int afc_mem_tracker_clear ( MemTracker * mem_tracker ) 
/*
@node afc_mem_tracker_clear

	         NAME: afc_mem_tracker_clear ( mem_tracker )  - Clears all stored data

	     SYNOPSIS: int afc_mem_tracker_clear ( MemTracker * mem_tracker)

	  DESCRIPTION: Use this command to clear all stored data in the current mem_tracker instance.
		       If the clear function has been set, items are also cleared from memory.

	        INPUT: - mem_tracker    - Pointer to a valid afc_mem_tracker instance.

	      RESULTS: should be AFC_ERR_NO_ERROR
	              
	     SEE ALSO: - afc_mem_tracker_delete()
	               
@endnode
*/
int afc_mem_tracker_clear ( MemTracker * am ) 
{
	unsigned int t;
	MemTData * d;
	unsigned int leak_size = 0;

	if ( am == NULL ) return ( AFC_LOG_FAST ( AFC_ERR_NULL_POINTER ) );
 
	if ( am->magic != AFC_MEM_TRACKER_MAGIC ) return ( AFC_LOG_FAST ( AFC_ERR_INVALID_POINTER ) );

	// printf ( "TRACKER CLEAR\n" );
	// afc_mem_tracker_internal_debug_dump ( am );

	for ( t = 0; t < am->num_items; t ++ )
	{
		d = am->mem [ t ];

		if ( __internal_afc_base->start_log_level >= AFC_LOG_WARNING ) 
			_afc_dprintf ( "WARNING: MemTracker: freeing %x (size: %d)\n", ( int ) d->data, d->size );

		leak_size += d->size;

		//if ( ( ( char * ) ( d->data ) ) [ 0 ] != 0 ) free ( d->data );
		free ( d->data );
		free ( d );
	}

	if ( ( __internal_afc_base->start_log_level >= AFC_LOG_WARNING )  && ( leak_size ) )
		_afc_dprintf ( "WARNING: MemTracker: Total Memory Leaked: %d bytes\n", leak_size );

	am->num_items = 0;
	am->current_pos = 0;

	return ( AFC_ERR_NO_ERROR );
}
// }}}
// {{{ void * afc_mem_tracker_malloc ( MemTracker * hm, size_t size )
/*
@node afc_mem_tracker_malloc

	         NAME: afc_mem_tracker_malloc ( mem_tracker, size )  - Allocs some memory

	     SYNOPSIS: void * afc_mem_tracker_add ( MemTracker * mem_tracker, size_t size )

	  DESCRIPTION: This function allocs a memory block and remembers it inside the tracker.

	        INPUT: - mem_tracker  - Pointer to a valid afc_mem_tracker class.
			- size		- Memory size to alloc (in bytes).

	      RESULTS: should be AFC_ERR_NO_ERROR

	     SEE ALSO: - afc_mem_tracker_free()
@endnode
*/
void * afc_mem_tracker_malloc ( MemTracker * mt, size_t size )
{
	MemTData * hd;
	void * mem;


	if ( ( mem = malloc ( size ) ) == NULL ) return ( NULL );

	if ( ( __internal_afc_base->start_log_level >= AFC_LOG_NOTICE )  && ( mt->show_mallocs ) )
		_afc_dprintf ( "NOTICE: MemTracker: alloc %x (%d)\n", ( int ) mem, ( int ) size );

	hd = ( MemTData * ) malloc ( sizeof ( MemTData ) );
	// printf ( "MEM: %x (%d)\n", ( int ) hd, sizeof ( MemTData ) );

	if ( hd == NULL ) 
	{
		free ( mem );
		return ( NULL );
	}

	hd->hash_value = ( unsigned long ) mem;
	hd->data       = mem;
	hd->size       = size;

	mt->mem [ mt->num_items ] = hd;         // use am->num_items (number of items in list)
	mt->current_pos = mt->num_items;        // Set current pos to am->num_items 

	mt->num_items ++;			// Increment num_items

	mt->is_sorted = FALSE;

	if ( mt->num_items == mt->max_items ) afc_mem_tracker_internal_double_array ( mt );

	// printf ( "MEM: %x - HD: %x\n", ( int ) mem, ( int ) hd );

	// afc_mem_tracker_internal_debug_dump ( mt );

	return ( mem );
}
// }}}
// {{{ int afc_mem_tracker_free ( MemTracker * hm, void * mem )
/*
@node afc_mem_tracker_free

	         NAME: afc_mem_tracker_free ( mem_tracker, mem )  - Frees a memory block

	     SYNOPSIS: int afc_mem_tracker_add ( MemTracker * mem_tracker, void * mem )

	  DESCRIPTION: This function frees a memory block releasing it to the system.

	        INPUT: - mem_tracker  - Pointer to a valid afc_mem_tracker class.
			- mem		- Memory block to be freed

	      RESULTS: should be AFC_ERR_NO_ERROR

	     SEE ALSO: - afc_mem_tracker_malloc()
@endnode
*/
void afc_mem_tracker_free ( MemTracker * hm, void * mem )
{
	MemTData * hd;

	if ( mem == NULL ) return;

	if ( ( __internal_afc_base->start_log_level >= AFC_LOG_NOTICE )  && ( hm->show_frees ) )
		_afc_dprintf ( "NOTICE: MemTracker: free %x\n", ( int ) mem );

	free ( mem );

	hd = afc_mem_tracker_internal_find ( hm, mem );
	if ( hd == NULL ) return;
	free ( hd );

	afc_mem_tracker_internal_del ( hm );
}
// }}}


/*
	INTERNAL FUNCTIONS
*/
// {{{ int afc_mem_tracker_internal_double_array ( MemTracker * am )
static int afc_mem_tracker_internal_double_array ( MemTracker * am )
{
	unsigned long int m = ( ( sizeof ( void * ) ) * ( am->max_items  * 2 ) );
	
	if ( ( am->mem = realloc ( am->mem, m ) ) == NULL ) return ( AFC_LOG_FAST ( AFC_ERR_NO_MEMORY ) );

	am->max_items = am->max_items * 2;
	
	return ( AFC_ERR_NO_ERROR );
}
// }}}
// {{{ afc_mem_tracker_internal_sort ( hd1, hd2 )
static int afc_mem_tracker_internal_sort ( const void * hd1, const void * hd2 )
{
        MemTData * h1 = ( MemTData * ) ( * ( MemTData * *) hd1 );
        MemTData * h2 = ( MemTData *) ( * ( MemTData * *) hd2 );

        unsigned long int v1 = h1->hash_value;
        unsigned long int v2 = h2->hash_value;

	//printf ( "v1: %x - v2: %x\n", v1, v2 );
        return ( v1 > v2 );
}
// }}}
// {{{ afc_mem_tracker_internal_find ( mt, mem )
static void * afc_mem_tracker_internal_find ( MemTracker * mt, void * mem )
{
	MemTData * hd;
	int max, pos, min = 0;
	unsigned long hash_value = ( unsigned long ) mem;

	if ( mt->num_items <= 0 ) return ( NULL );

	if ( mt->is_sorted == FALSE ) afc_mem_tracker_internal_do_sort ( mt, afc_mem_tracker_internal_sort );

	// printf ( "FREE MEM: %x\n", ( int ) mem );

	max = mt->num_items;
	pos = max / 2;

	while ( pos!=-1 )
	{
		hd = ( MemTData * ) afc_mem_tracker_internal_item ( mt, pos );
		if ( hd == NULL ) return ( NULL );

		if ( hd->hash_value == hash_value ) return ( hd );

		if ( hd->hash_value  < hash_value )
			min = pos +1;
		else 
			max = pos -1;

		if ( min > max ) return ( NULL );

		pos = ( max - min ) / 2 + min;
	}

	return ( NULL );
}
// }}}
// {{{ afc_mem_tracker_internal_do_sort ( am, comp )
static void * afc_mem_tracker_internal_do_sort ( MemTracker * am, int ( *comp ) ( const void *, const void * ) ) 
{
	if ( am->num_items == 1 )
	{
		am->is_sorted = TRUE;
		return ( am->mem [ 0 ] );
	}

	qsort ( am->mem, am->num_items, sizeof ( void * * ), comp );

	// afc_mem_tracker_internal_debug_dump ( am );

	am->current_pos = 0;
	am->is_sorted = TRUE;

	return ( am->mem [ 0 ] );
}
// }}}
// {{{ afc_mem_tracker_internal_del ( am )
static int afc_mem_tracker_internal_del ( MemTracker * am )
{
	void * dest;
	void * src;
	size_t size;

	if ( am->num_items == 0 ) return ( AFC_ERR_NO_ERROR );

	if ( am->current_pos == ( am->num_items - 1 ) ) 	// If I enter here, then I am pointing on the last element of the array
	{
		// printf ( "DELETING LAST\n" );
		am->num_items --;				// The array is resized -1
		am->current_pos = am->num_items -1;		// and we'll point (again!) on the last element 

		// afc_mem_tracker_internal_debug_dump ( am );
		return ( AFC_ERR_NO_ERROR );
	}

	// am->is_sorted = FALSE;

	dest = &am->mem [ am->current_pos ];
	src  = &am->mem [ am->current_pos + 1 ];

	size = ( am->num_items - ( am->current_pos ) ) * ( sizeof ( void * ) ) ;

	memcpy ( dest, src, size );	// Flawfinder: ignore

	am->num_items --;

	if ( am->current_pos >= am->num_items ) am->current_pos = am->num_items-1;

	return ( AFC_ERR_NO_ERROR );
}
// }}}
// {{{ afc_mem_tracker_internal_item ( am, item )
static void * afc_mem_tracker_internal_item ( MemTracker * am, unsigned long item )
{
	if ( ( item >= am->num_items ) || ( item < 0 ) ) return ( NULL );

	return ( am->mem [ am->current_pos = item ] );
}
// }}}

#ifdef TEST_CLASS
// {{{ TEST_CLASS
int main ()
{
	AFC * afc = afc_new ();
	MemTracker * mt = afc_mem_tracker_new ();
	void * m;

	afc_mem_tracker_malloc ( mt, 1000 );
	afc_mem_tracker_malloc ( mt, 1000 );

	m = afc_mem_tracker_malloc ( mt, 2000 );

	afc_mem_tracker_free ( mt, m );

	afc_mem_tracker_delete ( mt );
	afc_delete ( afc );

	return ( 0 );
}
// }}}
#endif
